/* Start and monitor SRE-http -- reloading if a crash is detected */

/* Begin USER Changeable Parameters */
/* Begin USER Changeable Parameters */

/* IP (name) address to monitor. Set IP_ADDRESS='' to use "own" address. */
ip_name=''

/* fully qualified location of GoServe. If '', use GOSERVE.EXE (i.e.; in
   current directory */
goserve=''

/* Filter to invoke. Choices include srefiltr.80, sreproxy.80, and srelite.80
The default is srefiltr.80 */
filter=''

/* Update frequency, in minutes. More frequent updates will mean less time between
"recoveries", but will increase server load. */
update=5

/* Log file -- events are written here. Files are relative to the current
   directory (typically, the goserve\srehttp directory)*/
logfile='startsre.log'

/* If more then this number of consectuive timeouts (or connection refusals)
   detected; then kill and restart GoServe. 
   To suppress this, set MAX_TIMEOUTS=100000000, or set TRY_RESTART=0 */
max_timeouts=4

/* If possibly bad problem detected, attempt to kill, then restart, GoServe.
   Values are:
        Try_Restart = 0         -- do not attempt to restart
        Try_Restart = 1         -- attempt to restart, and also clear macrospace
        Try_Restart = 2         -- attempt to restart, but do NOT clear macrospace  */
Try_Restart=1


/* Use this on the "first hit" to the server (i.e.; hit the main page)
   For example: first_hit='/index.htm'
   To use the default, set first_hit='' */
first_hit='/'

/* Use  this on other hits. If =0, then a simple !PING  is used */
otherhits=0

/* verboseness of output: 0=normal, 1=verbose */
verbose=0

/* HINT: STARTSRE works a bit better when the MAIN_SERVER parameter (in SREFILTR.80) is equal to:
         MAIN_SERVER=' '
*/

/* ********************** END user changable parameters **************/

if arg(1)='?' then do
say' This is the SRE-http monitor'
say ' It will start, and if necessary, restart, SRE-http'
say ' To use it, you may need to set a few parameters (at the top of STARTSRE.CMD).'
exit
end

call load       /*rxsock load */

say
say ' This is the SRE-http monitor. '
say ' It will start, and if necessary, restart, SRE-http'
say
say 'Notes: Server = 'IP_name ' & filter='filter
say '       Messages will be written to 'logfile
say

call lineout logfile,' ----- STARTSRE invoked at ' date('n') time('n')

if ip_name='' then ip_name=get_hostname()
gospid=0 ; timeouts=0
if filter='' then filter='srefiltr.80'
if goserve='' then goserve='GOSERVE'
if first_hit='' then first_hit='/!PING'

if otherhits=0 | otherhits='' then otherhits='/!PING?PING'


/* -------- is goserve currently running? */
gospid=get_goserve_process(goserve)

if gospid<0 then do
   say "Error: Multiple instances of GoServe detected (possibly running on multiple ports)."
   say  "     StartSRE is not designed for this. "
   call lineout logfile, "Error: Multiple instances of GoServe detected (possibly running on multiple ports)."
   call lineout logfile, "     StartSRE is not designed for this. "
end

/* no... so start it */
if gospid=0  then  gospid=start_goserve(goserve,filter,ip_name,tryrestart)
if gospid=0 then do
   say "ERROR: could not start GoServe! "
   call lineout logfile,' ERROR: could not start GoServe! '
   exit
end /* do */

say "  GoServe running as process: " gospid
call lineout logfile,"  GoServe running as process: " gospid

hitwith=first_hit
getshead=1

if verbose=1 then say "Checking "ip_name " with " first_hit
wow=doit(ip_name,first_hit,1)              /* hit it with a HEAD !PING */
if wow<0 then say "Problem: " wow
if wow>0 then do
  say "Server at " ip_name " is currently running.... "
  call lineout logfile,"Server at " ip_name " is currently running.... "
  getshead=0
  hitwith=otherhits
end
else do                 /* couldn't access server, so restart it */
  call restart_goserve
end /* do */

nsec=trunc(update*60)+1
nbads=0

/* ------ start monitoring .. . */

/* -3=could not connect, -2=could not recieve, -1=503 error response, 1= okay */

do forever              /* monitor .... */

  wow=doit(ip_name,hitwith,getshead)              /* hit it with a HEAD xxx */

  if wow=1 then do                      /* succdeess == nap and try again */
      say 'server okay at '||time('n')||'; will now wait for 'nsec " seconds"
      call syssleep nsec                /* the "update" # of seconds */
      hitwith=otherhits
      getshead=0
      nbads=0
      iterate
  end

  if try_restart<1  then do     /* some kind of error, just note it and try again: */
      say "Error Detected ("wow"), try again in 10 seconds "
      call lineout logfile, "Error Detected ("wow"), try again in 10 seconds "
      call syssleep 10
      hitwith='/!PING?PING'
      getshead=0
      iterate
  end /* do */

/* provbems, perhaps restart. */

   if wow=-3 then do             /* fatal - could not cnnect */
     call lineout logfile,date('n')' 'time('n') ' Server at ' ip_name ' is down... restarting'
     hitwith=first_hit
     getshead=1
     call lineout logfile
     call restart_goserve        /* might EXIT */
     nbads=0
     iterate                             /* and immediately hit it*/
  end

  if wow=-2 then do              /* timeout? could not recieve */
    nbads=nbads+1
    say time('n') ", server at " ip_name " timed out or connection refused "
    call lineout logfile,time('n') ", server at " ip_name " timed out or connection refused "
    call syssleep 5                                     /* wait a sec */
    nbads=nbads+1
 end
 if wow=-1 then do               /* 503 response */
     say time('n') ", server at " ip_name " is having problems "
     call lineout logfile, say time('n') ", server at " ip_name " is having problems "
     nbads=nbads+1
     call syssleep 5                     /* wait a sec */
 end
 if nbads>maxtimeout then do             /* restart after consecutive maxtimeout failures */
     call restart_goserve
     nbads=0
 end /* do */

end             /* of forever */


/**********/
/* restart goserve */
restart_goserve:
   if tryrestart=0 then do
       say "ERROR: could not access server."
        call lineout logfile,"ERROR: could not access server"
        exit
   end /* do */
   /* find goserve process.  Pid if found, 0 if not running, -n if more then one instance */
   koo=get_goserve_process(goserve)
   if koo>0 then do
      foo=kill_goserve(koo)
   end
   gospid=start_goserve(goserve,filter,ip_name,tryrestart)
   if gospid<=0 then do
      say "ERROR: problem restarting GoServe "
      call lineout logfile,'ERROR: problem restarting GoServe '
      exit
   end /* do */
   say "  GoServe running as process: " gospid
  call lineout logfile,"  GoServe running as process: " gospid
  return 1

/**************/
/* start goserve and return it's pid */
start_goserve:procedure
parse upper arg toget,filter,ip_name,

if toget='' then toget='GOSERVE.EXE'
if pos('.',toget)=0 then toget=toget'.EXE'

  if try_restart=1 then do
     foo=macroclear()
     if foo=1 then 
         say "      Macrospace has been cleared."
     else
         say "      Could NOT clear macrospace"
  end

 say "Starting " filter " on " ip_name
  address cmd '@START "GoServe-SRE-http WWW server"  /MIN  ' GOSERVE '  http Filter 'filter ' reuse '
  call syssleep 3         /* wait a second */

gpid=get_goserve_process(toget)


return gpid

/******************************/
/* find goserve process.  Pid if found, 0 if not running, -n if more then one instance */
get_goserve_process:procedure
parse upper arg toget
if toget='' then toget='GOSERVE.EXE'
if pos('.',toget)=0 then toget=toget'.EXE'

ii=0
foo=dospidlist('pidstem','namestem','f1','f2')
do mm=1 to pidstem.0
     if pos(toget,upper(namestem.mm))>0 then do
        ii=ii+1
        gospid=pidstem.mm
    end
end /* do */
if ii=0 then return 0 
if ii=1 then return gospid 
if ii>1 then return -ii


/**************************/
/* kill goserve on process gospid */
kill_goserve:procedure
parse arg gospid


foo=doskillprocess(gospid)
if foo=0 then do  
     say "ERROR: unable to flag process for termination"
     exit
end /* do */
do mm=1 to 100
            foo=dosprocinfo('N',gospid)
            if foo='' then do
               say "   .... GoServe successfully killed!"
               leave
            end
            call syssleep 1
            return 1
end /* do */
say " *** UNABLE to kill GoServe after 100 seconds"
exit


/* ------------------------------------------------------------------- */
/* send a HEAD !PING request */
/* -3=could not connect, -2=could not recieve, -1=503 error response, 1= okay */

doit:procedure expose gospid  logfile verbose
parse arg server,amess,getshead

crlf    ='0d0a'x                        /* constants */
opts="" ;upwd=""

family  ='AF_INET'
httpport=80

rc=sockgethostbyname(server, "serv.0")  /* get dotaddress of server */
if rc=0 then do; say ' ERROR: Unable to resolve "'server'"'; return -1; end
dotserver=serv.0addr                    /* .. */
if verbose=1 then say " ..... address = " dotserver

gosaddr.0family=family                  /* set up address */
gosaddr.0port  =httpport
gosaddr.0addr  =dotserver

gosock = SockSocket(family, "SOCK_STREAM", "IPPROTO_TCP")
if verbose=1 then say " ..... socket =  " gosock 
message='HEAD  'amess ' HTTP/1.0'crlf'HOST:'server||crlf

message=message||crlf

got=''
rc = SockConnect(gosock,"gosaddr.0")
if rc<0 then return -3    /* could not connect error */ 

if verbose=1 then say " .... sending mesage "
rc = SockSend(gosock, message)
/* Now wait for the response */
totrc=0
if verbose=1 then say " .... waiting for response "

rc=sockin(gosock,20,1000,,verbose)
if abbrev(rc,'#SOCKIN')>0 then return -2

/*  rc = SockRecv(gosock, "response", 1500) */

got=got||response
if rc<0 then return -2          /* timeout error */

rc = SockClose(gosock)

parse var got ih nval nmess  '0d0a'x .
nval=strip(nval)
if nval=503 then return -1                      /* server reports a problem */

/* if here, okay */
if getshead=1 then do   /* find the "server:" response header */
   do forever
      if got='' then leave
      parse var got a1 '0d0a'x got
      a1a=strip(translate(a1))
      if abbrev(a1a,'SERVER:') then do
         say ' >>> 'a1
         call lineout logfile,' >>> 'a1
         leave
      end /* do */
   end /* do */
end /* do */

return 1

/* get the hostname (numriec.numeric) for this machine */
get_hostname: procedure
    do queued(); pull .; end                   /* flush */
    address cmd '@hostname'  '| rxqueue'    
    parse pull hostname  
    return hostname



/* --- Load the function library, if necessary --- */
load:
if \RxFuncQuery("SockLoadFuncs") then return      /* already there */
call RxFuncAdd "SockLoadFuncs","rxSock","SockLoadFuncs"
call SockLoadFuncs

/* Load up advanced REXX functions */
foo=rxfuncquery('sysloadfuncs')
if foo=1 then do
  call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
  call SysLoadFuncs
end

foo=rxfuncquery('rexxlibregister')
if foo=1 then do
 call rxfuncadd 'rexxlibregister','rexxlib', 'rexxlibregister'
 call rexxlibregister
end

return

/**************************/
/* SOCKIN: a replacement for sockrecv. 
  call as
      stuff=sockin(socket,timeout,maxlen,timeoutmess,verbose)
  where:
     socket == a socket that's been established using sockconnect
    timeout == a timeout value in seconds
    maxlen  == maximum length of message to recieve
               If not specified, then no maximum is imposed
  timeoutmess == Prefix for "error" and "timeout" message.
                 If not specified, "#SOCKIN: " is used as a prefix
                 For example: #SOCKIN: timeout " will be returned if no response
                 was recieved in timeout seconds.
  verbose  == If 1, then report status messages. 
              If not specified, then do NOT report status messages 
and
   stuff = the contents returned from the server (up to maxlen characters)
           or an error message (starting with the timeoutmess)

Note: timeout refers to maximum seconds between "sockrecv" requests. 
      It does NOT refer to total length of time required to recieve a message.
      Thus, a number of medium length delays (say, a few seconds required
      to complete each of several sockrecv requests) will NOT cause a timeout 
      (in other words, the timeout counter is reset upon each successful 
      completion of a 1000 byte sockrecv request).

*/

SOCKIN:PROCEDURE 
PARSE Arg socket,timeout,maxlen,timeoutmess,verbose

if maxlen=0 | maxlen='' then maxlen=100000000
if timeoutmess=''  then timeoutmess='#SOCKIN:'
if timeout='' then timeout=0.1

if verbose=1 then say "SockIn: will wait for " timeout " seconds ...."

if Sockioctl(socket,'FIONBIO',1) = -1 then /* switch to nonblocking mode */
   Return timeoutmess||'crashed in ioctl 'errno

ict=0
ok=0
incoming=''

Do While TimeOut > 0

    res=Sockrecv(socket,'data',1000) 
    if res=-1 then do                           /* error condition ? */

      If errno <> 'EWOULDBLOCK' THEN       /* real crash ? */
            Return timeoutmess||'crashed in sockrecv 'errno /* yes */

      /* not-fatal,no-data-available-condition:  errno = EWOULDBLOCK & sockrecv returned -1 */

      ict=ict+1
      if verbose=1 then say 'SockIn: Waiting ('ict')...'
      Call SysSleep 1              /* relinquish the CPU for one second, this is from REXXUTIL */

      TimeOut = TimeOut - 1; /* count down my timer */
      Iterate; /* loop again */
    End; /* if sockrecv = -1 */
    if res=0 then do
        ok=1 ; leave            /* got end of message, so exit this do loop*/
    end 
    if res<0 then do
         return timeoutmess||" Error in sockrecv " rc
    end
    incoming=incoming||data
    if verbose=1 then say "SockIn: total data recieved: "length(incoming) 
    if length(incoming)>maxlen then do
       ok=2
       leave
    end 
    ict=0               /* reset counter */
End /* do while timeout > 0 */

/* here we are timed out, or got entire message */
if ok=1  then do
   If sockioctl(socket,'FIONBIO',0) = -1 then do /* switch to blocking mode */
       if verbose=1 then say 'SockIn: ioctl error on switch to blocking mode: ' errno
   end
   return incoming                              /* success! */
end
if ok=2  then do
   if sockioctl(socket,'FIONBIO',0) = -1 then do  /* switch to blocking mode */
      if verbose=1 then say 'SockIn: ioctl error on switch to blocking mode: ' errno
   end
   Return timeoutmess||' message length exceeds maximum of 'maxlen
end

Return timeoutmess||' timeout '




